import pandas as pd
import altair as alt
import numpy as np

def plot_num_dist(data):
    """
    Creates a chart of histograms for all numeric features in a data set.

    Parameters
    ----------
    data : pandas.DataFrame
        An object containing the dataset of interest

    Returns
    -------
    altair.Chart
        Altair chart object

    Examples
    --------
    >>> plot_num_dist(dataset)

    """
    if not isinstance(data, pd.DataFrame):
        raise TypeError('dataset must be a pandas DataFrame')
    
    # Get list of numeric features
    numeric_features = data.select_dtypes(include=np.number).columns.to_list()

    # Create plot
    plot = alt.Chart(data).mark_bar().encode(
        alt.X(alt.repeat()).type('quantitative').bin(maxbins=20),
        alt.Y('count()', stack=None).title('Frequency of Occurrence')
    ).properties(
        width = 125,
        height = 125
    ).repeat(
        numeric_features,
        columns = 3
    )
    return plot





import numpy as np
import pandas as pd

def generate_financial_metrics(random_returns_df, annual_risk_free=0.00):
    """
    Calculate financial metrics based on a DataFrame of random returns on time series.

    Parameters
    ----------
    random_returns_df : pandas.DataFrame
        DataFrame containing returns of stocks on time series.

    annual_risk_free : float, default 0.00.
        Annualized risk-free rate as a decimal (e.g., 0.02 for 2%).
        
    Returns
    -------
    pandas.DataFrame
        A DataFrame containing financial metrics including count (number of returns), 
        total return, annualized return, annualized volatility, and Sharpe ratio for each
        stock return series

    Examples
    --------
    >>> random_returns_df = pd.DataFrame([
    {'index': '2024-01-01 00:00:00', 'stock_1': -0.002286, 'stock_2': 0.002545, 'stock_3': 0.002997},
    {'index': '2024-01-01 01:00:00', 'stock_1': 0.003045, 'stock_2': 0.002913, 'stock_3': -0.000855},
    {'index': '2024-01-01 02:00:00', 'stock_1': -0.001142, 'stock_2': 0.000635, 'stock_3': -0.002592},
    {'index': '2024-01-01 03:00:00', 'stock_1': 0.000971, 'stock_2': -0.000168, 'stock_3': 0.000074},
    {'index': '2024-01-01 04:00:00', 'stock_1': -0.000392, 'stock_2': 0.002252, 'stock_3': -0.000342}
    ])
                                                   
    >>> generate_financial_metrics(random_returns_df, annual_risk_free=0.02)
            	count	total_return	annual_return	annual_volatility	sharpe_ratio
    stock_1 	5	    0.000196	    0.343184	    0.192332	        1.680339
    stock_2 	5	    0.008177	    14.325234	    0.124650	        114.763058
    stock_3 	5	    -0.000718	    -1.258381	    0.189806	        -6.735214
    """

    if random_returns_df.empty:
        raise ValueError("Random returns DataFrame is empty.")
    
    if not isinstance(random_returns_df.index, pd.DatetimeIndex):
        raise ValueError("Index must be a pd.DatetimeIndex.")  

    if random_returns_df.isnull().values.any():
        raise ValueError("Random returns DataFrame contains missing values.")

    if not isinstance(annual_risk_free, (int, float)):
        raise ValueError("Annual risk-free rate must be numeric.")

    freq = pd.infer_freq(random_returns_df.index)
    
    if freq == 'D':
        periods_per_year = 365
    elif freq == 'H':
        periods_per_year = 365 * 24
    elif freq == 'min':
        periods_per_year = 365 * 24 * 60
    else:
        raise ValueError("Invalid frequency. Frequency should be either 'D' for daily, 'H' for hourly, or 'min' for minutely.") 
    
    metrics_df = pd.DataFrame({
        'count': random_returns_df.count(),
        'total_return': random_returns_df.sum(),
        'annual_return': random_returns_df.mean() * periods_per_year,
        'annual_volatility': random_returns_df.std() * np.sqrt(periods_per_year),
    })
    metrics_df['sharpe_ratio'] = (metrics_df['annual_return'] - annual_risk_free) / metrics_df['annual_volatility']
    
    return metrics_df



import pandas as pd
import numpy as np

def generate_return_series(
    expected_annual_return, 
    annual_volatility, 
    n_rows=365, 
    freq='D', 
    num_series=1, 
    dist='normal', 
    random_state=524,
    start_date='2024-01-01'):
    """
    Generate a DataFrame with independent time series of returns.

    Parameters
    ----------
    expected_annual_return : float
        Expected annualized return as a decimal (e.g., 0.05 for 5%).

    annual_volatility : float
        Annualized volatility as a decimal (e.g., 0.2 for 20%).

    n_rows : int, default 365 
        Number of days, hours, or minutes (rows) to generate.

    freq : {'D', 'H', 'min'}, default 'D'
        Frequency of returns ('D' for daily, 'H' for hourly, 'min' for minute).

    num_series : int, default 1
        Number of independent time series (columns) to generate.

    dist : {'normal', 'lognormal'}, default 'normal'
        Type of return distribution. Only support Normal and Log-normal distribution.

    random_state : int, default 524
        Seed for numpy random number generation.

    start_date : str, default '2024-01-01'
        Start date for the series in the format 'YYYY-MM-DD'. 

    Returns
    -------
    pandas.DataFrame
        A DataFrame with columns as independent return time series.

    Examples
    --------
    >>> generate_return_series(0.05, 0.2, n_rows=3, freq='D', num_series=3, dist='normal', random_state=123, start_date='2024-01-01')
                series_1  series_2  series_3
    2024-01-01 -0.012345  0.016767  0.022222
    2024-01-02  0.043215  0.019105 -0.005555
    2024-01-03 -0.001011  0.003333 -0.011111
    """
    np.random.seed(random_state)
    if annual_volatility < 0:
        raise ValueError("annual_volatility < 0.")

    if freq == 'D':
        periods_per_year = 365
    elif freq == 'H':
        periods_per_year = 365 * 24
    elif freq == 'min':
        periods_per_year = 365 * 24 * 60
    else:
        raise ValueError("Invalid frequency. Use 'D' for day, 'H' for hour, 'min' for minute.") 

    num_periods = int(periods_per_year) * num_series

    if dist == 'normal':
        returns = np.random.normal(
            expected_annual_return / periods_per_year,
            annual_volatility / np.sqrt(periods_per_year),
            (n_rows, num_series)
        )
    elif dist == 'lognormal':
        mu = np.log(1 + expected_annual_return) - 0.5 * (annual_volatility ** 2) / periods_per_year
        sigma = annual_volatility / np.sqrt(periods_per_year)
        returns = np.random.lognormal(mu, sigma, (n_rows, num_series))
    else:
        raise ValueError("Invalid distribution. Use 'normal' or 'lognormal'.")

    datetime_range = pd.date_range(start=start_date, periods=n_rows, freq=freq)
    
    return pd.DataFrame(returns, index=datetime_range, columns=[f'series_{i+1}' for i in range(num_series)])



import altair as alt
import pandas as pd

def plot_missing_vals(data):
    """
    Plot binary heatmap chart ('True' where there is a missing value and 'False' where there is not)
    to display missing values for all numeric features in the dataset.

    This function is derived from UBC Master of Data Science course DSCI 531 Lecture 4 notes by Joel Ostblom 
    (https://pages.github.ubc.ca/MDS-2023-24/DSCI_531_viz-1_students/lectures/4-eda.html)

    Parameters
    ----------
    data : pandas.DataFrame
        The dataset.

    Returns
    -------
    altair.Chart
        Altair chart object.

    Examples
    --------
    >>> plot_missing_vals(data)
        
    """
    

    if not isinstance(data, pd.DataFrame):
        raise ValueError("Data must be a pandas DataFrame")
    
    missing_vals_plot = alt.Chart(
        data.isna().reset_index().melt(
            id_vars='index'
        )
    ).mark_rect().encode(
        alt.X('index:O').axis(None),
        alt.Y('variable').title("Features"),
        alt.Color('value').title('Missing Values'),
        alt.Stroke('value')  
    ).properties(
        width=data.shape[0]
    )

    return missing_vals_plot
